import os.path as osp
from typing import Optional, List
import io 

import tarfile
import numpy as np
import cv2 

import mmengine.fileio as fileio
from mmengine.registry import TRANSFORMS
from mmcv.transforms import LoadImageFromFile
from mmcv.transforms import LoadAnnotations as MMCV_LoadAnnotations

from rasterio.io import MemoryFile 


def bandsfrombytes(content: bytes,
                band_ids: Optional[List[int]] = None) -> dict:
    """Read multiple bands from bytes.
    Args:
        content (bytes): raster bytes got from files or other streams.
        band_ids (list of int): band ids to be selected
    Returns:
        dict: Loaded bands as dict map with band keys and band array.
    Examples:
        >>> tif_path = '/path/to/img.tif'
        >>> with open(tif_path, 'rb') as f:
        >>>     tif_buff = f.read()
        >>> bands = mmcv.bandsfrombytes(tif_buff, band_ids=[0, 1, 2])
    """
    with MemoryFile(content) as memfile:
        r = memfile.open()
        bands = r.read()
        metadata = r.profile
        r.close()
        
        if not band_ids:
            return bands, metadata
        else:
            return bands[band_ids], metadata


def bandsfromtarbytes(content: bytes,
                band_substrings: List[str]) -> dict:
    """Read bands from tar bytes.
    Args:
        content (bytes): raster or tar bytes got from files or other streams.
        band_substrings (list of str): bands substring present in file names inside tar
    Returns:
        dict: Loaded bands as dict map with band keys and band array.
    Examples:
        >>> tar_path = '/path/to/img.tar'
        >>> with optn(tar_path, 'rb') as f:
        >>>     tar_buff = f.read()
        >>> img = mmcv.bandsfrombytes(tar_buff, band_substrings=['B01', 'B02', 'B03'])
    """
    with io.BytesIO(content) as buff:
        tar = tarfile.open(fileobj=buff, mode='r')
        bands = {}
        metadata = {}
        for member in tar.getmembers():
            if 'tif' in member.name:
                with MemoryFile(tar.extractfile(member.name).read()) as memfile:
                    with memfile.open() as dataset:
                        bands[member.name] = dataset.read()
                        metadata[member.name] = dataset.profile

        out_bands = {}
        out_metadata = {}
        for band_str in band_substrings:
            for k in bands.keys():
                if band_str in k:
                    out_bands[band_str] = bands[k]
                    out_metadata[band_str] = metadata[k]
        
        tar.close()

    return out_bands, out_metadata


@TRANSFORMS.register_module()
class LoadBandsFromFile(LoadImageFromFile):
    """Load bands from file.
    Required keys are "img_prefix" and "img_info" (a dict that must contain the
    key "filename"). Added or updated keys are "filename", "img", "img_shape",
    "ori_shape" (same as `img_shape`) and "img_norm_cfg" (means=0 and stds=1).
    "img_meta" key contains raster profile metadata.
    Args:
        to_float32 (bool): Whether to convert the loaded image to a float32
            numpy array. If set to False, the loaded image is an uint8 array.
            Defaults to False.
        band_ids (list(int)): List of bands to subsample from loaded band files.
            Defaults to all bands.
        file_client_args (dict): Arguments to instantiate a FileClient.
            See :class:`mmcv.fileio.FileClient` for details.
            Defaults to ``dict(backend='disk')``.
    """

    def __init__(self,
                 to_float32=False,
                 gray2rgb=False,
                 band_ids=None,
                 file_client_args=dict(backend='disk')):
        self.band_ids = band_ids
        self.to_float32 = to_float32
        self.file_client_args = file_client_args.copy()
        self.file_client = None
        self.gray2rgb = gray2rgb

    def transform(self, results: dict):
        filename = results['img_path']
        if self.file_client_args is not None:
            file_client = fileio.FileClient.infer_client(
                self.file_client_args, filename)
            band_bytes = file_client.get(filename)
        else:
            band_bytes = fileio.get(
                filename, backend_args=self.backend_args)
        
        bands, metadata = bandsfrombytes(band_bytes, band_ids=self.band_ids)
        if self.gray2rgb or bands.shape[0] == 1:
            # assert bands.shape[0] == 1, "gray2rgb is only for grayscale images"
            bands = np.asarray([bands[0], bands[0], bands[0]])
        bands = bands.transpose(1, 2, 0)

        if "slice_x" in results:
            bands = self._make_patch(bands, results)
        if self.to_float32:
            bands = bands.astype(np.float32)
                
        results["img"] = bands
        results["img_shape"] = bands.shape[:2]
        results["ori_shape"] = bands.shape[:2]
        results["img_meta"] = metadata

        num_channels = 1 if len(bands.shape) < 3 else bands.shape[2]

        results['img_norm_cfg'] = dict(
                mean=np.zeros(num_channels, dtype=np.float32),
                std=np.ones(num_channels, dtype=np.float32),
                to_rgb=False)

        return results


    def _make_patch(self, bands, img_info):
        slice_x = img_info["slice_x"]
        slice_y = img_info["slice_y"]
        slice_w = img_info["slice_w"]
        slice_h = img_info["slice_h"]
        return bands[slice_y:slice_y+slice_h, slice_x:slice_x+slice_w, :].copy()


@TRANSFORMS.register_module()
class LoadVariableSizedBandsFromFile(LoadImageFromFile):
    """Load variable sized bands from tar file.
    Required keys are "img_prefix" and "img_info" (a dict that must contain the
    key "filename"). Added or updated keys are "filename", "img", "img_shape",
    "ori_shape" (same as `img_shape`) and "img_norm_cfg" (means=0 and stds=1).
    "img_{BAND_ID}" contains raster profile for band with band id.
    Args:
        band_substrings (list(str)): Band substrings that must match band file names present in the tar file.
            Required argument.
        to_float32 (bool): Whether to convert the loaded image to a float32
            numpy array. If set to False, the loaded image is an uint8 array.
            Defaults to False.
        file_client_args (dict): Arguments to instantiate a FileClient.
            See :class:`mmcv.fileio.FileClient` for details.
            Defaults to ``dict(backend='disk')``.
    """

    def __init__(self,
                 band_substrings,
                 to_float32=False,
                 file_client_args=dict(backend='disk')):
        self.band_substrings = band_substrings
        self.to_float32 = to_float32
        self.file_client_args = file_client_args.copy()
        self.file_client = None

    def transform(self, results: dict):
        filename = results['img_path']
        if self.file_client is None:
            self.file_client = fileio.FileClient.infer_client(
                    self.file_client_args, filename)
            
        bands_bytes = self.file_client.get(filename)
        bands, metadata = bandsfromtarbytes(bands_bytes, band_substrings=self.band_substrings)
        band_shapes = {}
        if self.to_float32:
            for band_id in bands.keys():
                bands[band_id] = bands[band_id].astype(np.float32)
                band_shapes[band_id] = bands[band_id].shape 

        for band_id in bands.keys():
            results[band_id] = bands[band_id]
            results[band_id + '-meta'] = metadata[band_id]
        results['band_shapes'] = band_shapes
        results['ori_shapes'] = band_shapes
        num_channels = len(self.band_substrings)
        results['bands_norm_cfg'] = dict(
            mean=np.zeros(num_channels, dtype=np.float32),
            std=np.ones(num_channels, dtype=np.float32),
            to_rgb=False)

        return results

    def __repr__(self):
        repr_str = (f'{self.__class__.__name__}('
                    f'to_float32={self.to_float32}, '
                    f"band_substrings='{self.band_substrings}', "
                    f'file_client_args={self.file_client_args})')
        return repr_str

@TRANSFORMS.register_module()
class LoadMasks(MMCV_LoadAnnotations):
    """Load annotations for semantic segmentation provided by GeoEngine dataset.

    The annotation format is as the following:

    .. code-block:: python

        {
            # Filename of semantic segmentation ground truth file.
            'seg_map_path': 'a/b/c'
        }

    After this module, the annotation has been changed to the format below:

    .. code-block:: python

        {
            # in str
            'seg_fields': List
             # In uint8 type.
            'gt_seg_map': np.ndarray (H, W)
        }

    Required Keys:

    - seg_map_path (str): Path of semantic segmentation ground truth file.

    Added Keys:

    - seg_fields (List)
    - gt_seg_map (np.uint8)

    Args:
        reduce_zero_label (bool, optional): Whether reduce all label value
            by 1. Usually used for datasets where 0 is background label.
            Defaults to None.
        imdecode_backend (str): The image decoding backend type. The backend
            argument for :func:``mmcv.imfrombytes``.
            See :fun:``mmcv.imfrombytes`` for details.
            Defaults to 'pillow'.
        backend_args (dict): Arguments to instantiate a file backend.
            See https://mmengine.readthedocs.io/en/latest/api/fileio.htm
            for details. Defaults to None.
            Notes: mmcv>=2.0.0rc4, mmengine>=0.2.0 required.
    """

    def __init__(
        self,
        reduce_zero_label=None,
        backend_args=None,
        imdecode_backend='pillow',
    ) -> None:
        super().__init__(
            with_bbox=False,
            with_label=False,
            with_seg=True,
            with_keypoints=False,
            imdecode_backend=imdecode_backend,
            backend_args=backend_args)
        self.reduce_zero_label = reduce_zero_label
        if self.reduce_zero_label is not None:
            warnings.warn('`reduce_zero_label` will be deprecated, '
                          'if you would like to ignore the zero label, please '
                          'set `reduce_zero_label=True` when dataset '
                          'initialized')
        self.imdecode_backend = imdecode_backend

    def _load_seg_map(self, results: dict) -> None:
        """Private function to load semantic segmentation annotations.

        Args:
            results (dict): Result dict from :obj:``mmcv.BaseDataset``.

        Returns:
            dict: The dict contains loaded semantic segmentation annotations.
        """
        if self.reduce_zero_label is None:
            self.reduce_zero_label = results['reduce_zero_label']
        assert self.reduce_zero_label == results['reduce_zero_label'], \
            'Initialize dataset with `reduce_zero_label` as ' \
            f'{results["reduce_zero_label"]} but when load annotation ' \
            f'the `reduce_zero_label` is {self.reduce_zero_label}'
        h, w = results['slice_h'], results['slice_w']
        gt_semantic_seg = np.zeros((h, w), dtype=np.uint8)
        for annotation in results["instances"]:
            for poly in annotation['mask']:
                if len(poly):
                    pts = np.asarray(poly).reshape(-1, 1, 2).astype(np.int32)[:-1]
                    label = annotation['bbox_label'] + 1
                    gt_semantic_seg = cv2.fillPoly(gt_semantic_seg, pts=[pts], color=label)
        if self.reduce_zero_label:
            # avoid using underflow conversion
            gt_semantic_seg[gt_semantic_seg == 0] = 255
            gt_semantic_seg = gt_semantic_seg - 1
            gt_semantic_seg[gt_semantic_seg == 254] = 255

        results['gt_seg_map'] = gt_semantic_seg
        results['seg_fields'].append('gt_seg_map')

        return results

    def __repr__(self) -> str:
        repr_str = self.__class__.__name__
        repr_str += f'(reduce_zero_label={self.reduce_zero_label}, '
        repr_str += f"imdecode_backend='{self.imdecode_backend}', "
        repr_str += f'backend_args={self.backend_args})'
        return repr_str


import base64

import torch
from torchvision.ops import nms
import numpy as np
import rasterio as rio
import cv2

from mmengine.config import Config
from mmengine.runner import Runner
from shapely.geometry import box
from mmdet.structures import DetDataSample


def rio2patches(img_r, cfg):
    """this function read a rio image and creates patches based on cfg.window and cfg.window.
    return list of patches (patches x window_height x window_width) and list of location of patches (x1, y1, x2, y2).
    Make sure every pixel of image is included in patches. We cannot ignore left and bottom sides.
    Duplicates are okay.
    """
    window_height, window_width = cfg.window if "window" in cfg else (512, 512)

    img_height, img_width = img_r.shape[1], img_r.shape[2]

    if window_height > img_height:
        window_height = img_height

    if window_width > img_width:
        window_width = img_width

    locs = []

    for i in range(0, img_height, window_height):
        for j in range(0, img_width, window_width):
            if img_height < i + window_height:
                y_start = img_height - window_height
                y_end = img_height
            else:
                y_start = i
                y_end = i + window_height

            if img_width < j + window_width:
                x_start = img_width - window_width
                x_end = img_width
            else:
                x_start = j
                x_end = j + window_width

            locs.append([x_start, x_end, y_start, y_end])

    return locs


def merge_bboxes(bboxes, scores, labels, location_of_patches):
    """This should be easy to merge, the bboxes locations will be localized
    to window_height and window_width, just translate them to original image
    location using location_of_patches. Once all bboxes are merged.
    Use apply non maximal supression, this will remove overlapping boxes area
    due to overlapping regions in patches."""

    flattened_bboxes = []
    flattened_scores = []
    flattened_labels = []

    for i, patch_bbox in enumerate(bboxes):
        x_start, x_end, y_start, y_end = location_of_patches[i]
        for bbox in patch_bbox:
            bbox[0] += x_start
            bbox[1] += y_start
            bbox[2] += x_start
            bbox[3] += y_start
            flattened_bboxes.append(bbox)
        flattened_scores += scores[i]
        flattened_labels += labels[i]

    if flattened_bboxes:
        keep_idx = nms(
            torch.tensor(flattened_bboxes),
            torch.tensor(flattened_scores),
            iou_threshold=0.5,
        )

        merged_bboxes = np.array(flattened_bboxes)[list(keep_idx.numpy())]
        merged_scores = np.array(flattened_scores)[list(keep_idx.numpy())]
        merged_labels = np.array(flattened_labels)[list(keep_idx.numpy())]

        return merged_bboxes, merged_scores, merged_labels

    return [], [], []


class Inference:
    def __init__(
        self,
        cfg_string,
        weight_file,
        GS_ACCESS_KEY_ID=None,
        GS_SECRET_ACCESS_KEY=None,
        logger=None,
    ):
        cfg = Config.fromstring(cfg_string, file_format=".py")
        cfg.launcher = "none"
        runner = Runner.from_cfg(cfg)

        checkpoint = torch.load(weight_file)
        runner.model.load_state_dict(checkpoint["state_dict"])
        _ = runner.model.eval()

        self.cfg = cfg
        self.runner = runner
        self.GS_ACCESS_KEY_ID = GS_ACCESS_KEY_ID
        self.GS_SECRET_ACCESS_KEY = GS_SECRET_ACCESS_KEY
        self.logger = logger

    def read_raster(self, tif_url):
        try:
            with rio.Env(
                GS_ACCESS_KEY_ID=self.GS_ACCESS_KEY_ID,
                GS_SECRET_ACCESS_KEY=self.GS_SECRET_ACCESS_KEY,
            ):
                r = rio.open(tif_url)
                img = r.read()
            self.logger.info(f"Reading raster {tif_url}")
            return r, img
        except:
            if self.logger:
                self.logger.error(f"Error in reading raster {tif_url}")

    def segmentation(self, tif_path):
        r, img = self.read_raster(tif_path)
        location_of_patches = rio2patches(img, self.cfg)

        img_height, img_width = img.shape[1], img.shape[2]

        full_mask = np.zeros((img_height, img_width), dtype=np.uint8)

        if self.logger:
            self.logger.info(
                f"Total number of patches for image {tif_path} are: {len(location_of_patches)}"
            )

        for i, locs in enumerate(location_of_patches):
            x_start, x_end, y_start, y_end = locs

            patch = img[:, y_start:y_end, x_start:x_end]
            model_input = {"inputs": [torch.from_numpy(patch)]}

            result = self.runner.model.test_step(model_input)
            pred_patch = result[0].seg_logits.data.detach().cpu().numpy()
            mask_patch = np.argmax(pred_patch, axis=0)
            mask_patch = np.array(mask_patch, dtype=np.uint8)
            full_mask[y_start:y_end, x_start:x_end] = mask_patch

        patches = []

        is_success, buffer = cv2.imencode(".png", full_mask)
        img_base64 = base64.b64encode(buffer).decode()
        lbl_ids = np.unique(full_mask)
        if self.logger:
            self.logger.info(
                f"Labels for this image are: {lbl_ids}"
            )
        return img_base64, lbl_ids

    def detection(self, tif_path):
        r, img = self.read_raster(tif_path)
        location_of_patches = rio2patches(img, self.cfg)
        
        img_shape = self.cfg.window if "window" in self.cfg else (512, 512)

        if self.logger:
            self.logger.info(
                f"Total number of patches for image {tif_path} are: {len(location_of_patches)}"
            )

        bboxes, scores, labels = [], [], []
        for i, locs in enumerate(location_of_patches):
            x_start, x_end, y_start, y_end = locs
            patch = img[:, y_start:y_end, x_start:x_end]

            model_input = {
                "inputs": torch.from_numpy(patch).unsqueeze(0),
                "data_samples": [
                    DetDataSample(
                        metainfo={
                            "img_shape": img_shape,
                            "ori_shape": img_shape,
                            "scale_factor": (1, 1),
                        }
                    )
                ],
            }
            result = self.runner.model.test_step(model_input)[0]

            bboxes.append(
                list(result.pred_instances.bboxes.detach().cpu().numpy())
            )
            scores.append(
                list(result.pred_instances.scores.detach().cpu().numpy())
            )
            labels.append(
                list(result.pred_instances.labels.detach().cpu().numpy())
            )
        assert len(bboxes) == len(scores) == len(labels) == len(location_of_patches)

        try:
            merged_bboxes, merged_scores, merged_labels = merge_bboxes(
                bboxes, scores, labels, location_of_patches
            )
            geo_bboxes = []
            for bbox in merged_bboxes:
                x1, y1, x2, y2 = bbox
                xs, ys = rio.transform.xy(r.transform, (y1, y2), (x1, x2))
                poly = box(xs[0], ys[1], xs[1], ys[0])
                geo_bboxes.append([list(poly.exterior.coords)])

            if self.logger:
                self.logger.info(
                    f"Total number of geo boxes for image {tif_path} are: {len(geo_bboxes)}"
                )
        except Exception as e:
            geo_bboxes, merged_scores, merged_labels = [], [], []
            if self.logger:
                self.logger.error(e)

        return geo_bboxes, merged_scores, merged_labels

    def classification(self, tif_path):
        pass



import numpy as np 

import mmcv 

from mmcv.transforms import BaseTransform
from mmengine.registry import TRANSFORMS


@TRANSFORMS.register_module()
class ResizeAllToThisBand(BaseTransform):
    """Resize all bands to a given band size and make a tensor.

    Args:
        keys (list): band keys.
        band_to_use (str): band key to use for the destination shape.
        interpolation (str): Interpolation method. For "cv2" backend, accepted
            values are "nearest", "bilinear", "bicubic", "area", "lanczos". For
            "pillow" backend, accepted values are "nearest", "bilinear",
            "bicubic", "box", "lanczos", "hamming".
            More details can be found in `mmcv.image.geometric`.
        backend (str): The image resize backend type, accepted values are
            `cv2` and `pillow`. Default: `cv2`.
    """

    def __init__(self,
                 keys,
                 band_to_use,
                 interpolation='bilinear',
                 backend='cv2'):
        
        if backend not in ['cv2', 'pillow']:
            raise ValueError(f'backend: {backend} is not supported for resize.'
                             'Supported backends are "cv2", "pillow"')
        if backend == 'cv2':
            assert interpolation in ('nearest', 'bilinear', 'bicubic', 'area',
                                     'lanczos')
        else:
            assert interpolation in ('nearest', 'bilinear', 'bicubic', 'box',
                                     'lanczos', 'hamming')
        self.keys = keys
        self.band_to_use = band_to_use
        self.interpolation = interpolation
        self.backend = backend

    def _resize_bands(self, results):
        h, w = results[self.band_to_use].shape[1:]
        img = []
        for key in self.keys:
            band = results[key]
            img.append(mmcv.imresize(
                    band[0],
                    size=(w, h),
                    interpolation=self.interpolation,
                    return_scale=False,
                    backend=self.backend))
        img = np.asarray(img).transpose(1, 2, 0)
        results['img'] = img
        results['img_shape'] = img.shape[:2]
        results['ori_shape'] = img.shape[:2]

        for key in self.keys:
            del results[key]

    def transform(self, results):
        self._resize_bands(results)
        return results

    def __repr__(self):
        repr_str = self.__class__.__name__
        repr_str += f'(size={self.size}, '
        repr_str += f'interpolation={self.interpolation})'
        return repr_str


import json, copy
import numpy as np
from tqdm import tqdm

import pandas as pd 

from mmengine.registry import DATASETS
from torch.utils.data import Dataset
from typing import List, Union, Callable, Optional, Sequence

from mmengine.dataset import BaseDataset, Compose
from multiprocessing import Pool


@DATASETS.register_module()
class EngineCSV(BaseDataset):
    def __init__(self,
                 data_path: str,
                 raster_dir_path: str,
                 vector_dir_path: str,
                 split: str = "train",
                 class_title: str = None,
                 pipeline: List[Union[dict, Callable]] = [],
                 multilabel: bool = False,
                 test_mode: bool = False,
                 serialize_data: bool = True,
                 lazy_init: bool = False,
                 indices: Optional[Union[int, Sequence[int]]] = None,
                 max_refetch: int = 1000):
        self.raster_data_path = f"{data_path}/{raster_dir_path}"
        self.vector_data_path = f"{data_path}/{vector_dir_path}"

        self.split = split
        self.class_title = class_title
        self.multilabel = multilabel
        self.test_mode = test_mode
        self.serialize_data = serialize_data
        self._indices = indices

        self.pipeline = Compose(pipeline)
        self.max_refetch = max_refetch

        if not lazy_init:
            self.full_init()

    def full_init(self):
        """Load annotation file and set ``BaseDataset._fully_initialized`` to
        True."""
        super().full_init()

        #  To support the standard OpenMMLab 2.0 annotation format. Generate
        #  metainfo in internal format from standard metainfo format.
        if 'categories' in self._metainfo and 'classes' not in self._metainfo:
            categories = sorted(
                self._metainfo['categories'], key=lambda x: x['id'])
            self._metainfo['classes'] = tuple(
                [cat['category_name'] for cat in categories])
    
    def process_record(self, record):

        file_name = record["image:01"].split("/")[-1]
        file_path = f"{self.raster_data_path}/{file_name}"

        if self.multilabel:
            gt_label = []
            for x in record[self.class_title].split('\t'):
                gt_label.append(self.class_map[x])
        else:
            gt_label = self.class_map[record[self.class_title]]

        return {
            "img_path": file_path,
            "img_id": record["image-id"],
            "gt_label": gt_label
        }

    def load_data_list(self):
        meta_path = f"{self.vector_data_path}/metadata.json"
        fin = open(meta_path, "r")
        metadata = json.load(fin)
        fin.close()

        if not self.class_title:
            self.CLASSES = metadata["label:metadata"][0]["options"]
            self.class_title = metadata["label:metadata"][0]["title"].replace(" ", "-").lower()
            self.class_title_idx = 0
        else:
            for idx, question in enumerate(metadata["label:metadata"]):
                if question["title"] == self.class_title:
                    self.CLASSES = question["options"]
                    self.class_title = self.class_title.replace(" ", "-").lower()
                    self.class_title_idx = idx 

        self._metainfo = {
            "title": metadata["title"],
            "description": metadata["description"],
            "tags": metadata["tags"],
            "problemType": metadata["problemType"] if "problemType" in metadata else None,
            "question_title": metadata["label:metadata"][self.class_title_idx]["title"],
            "question_description": metadata["label:metadata"][self.class_title_idx]["description"],
            "classes": metadata["label:metadata"][self.class_title_idx]["options"]
        }

        self.class_map = {v:k for k,v in enumerate(metadata["label:metadata"][self.class_title_idx]["options"])}

        vec_path = f"{self.vector_data_path}/{metadata['dataset'][self.split]}"
        df = pd.read_csv(vec_path)
        records = df.to_dict('records')
        
        pool = Pool(16)
        data_list = list(tqdm(pool.imap(self.process_record, records), total=len(records)))
        pool.close()

        return data_list
